import torch
import numpy as np
import scipy.optimize as sciopt
torch.set_default_dtype(torch.float64)
_pyomo = False
try:
    import pyomo.environ as pyo
    _pyomo = True
except ImportError:
    print('Pyomo not found. Defaulting to gradient-based solver.')

def ferntala(W, normalized = True, reltol = 5e-3, delta = 0.05, rho = 1):
    """Computes the Fernique-Talagrand functional for a finite point set W
    on R^d with respect to the Euclidean distance. 
    - Input: the matrix W = (w_1,...,w_n) is d x n. 
             (optional) a flag for normalizing the functional (dividing by
                        the diameter of the set W (defaults to True)
             (optional) relative tolerance for the optimizer
             (optional) a step size delta > 0 for the optimizer
    - Output: a single positive value
    """
   
    n = W.size(1)
    # Use professional solver if problem is sufficiently small
    if (n <= 20) and _pyomo:
        return ft_pyomo(W)
    
    # Compute Euclidean distance matrix
    norms = torch.sum(torch.mul(W, W), 0).unsqueeze(0).repeat(n,1)
    distances = torch.sqrt(torch.abs(norms + norms.T - 2*W.T.mm(W)))
    # Normalize by max
    if normalized:
        if rho is None:
            distances = distances / torch.max(distances)
        else:
            distances = torch.clamp(distances, 0.0, rho) / rho
    
    # Compute nearest-neighbour matrices
    distances, neighbours = torch.sort(distances, 1)
    distances = distances[:,1:n] - distances[:,0:n-1]
    neighbours = neighbours[:,:-1]
    
    # Optimize using gradient-based solver
    module = FerniqueTalagrandModule(distances, neighbours)
    optimizer = torch.optim.SGD(module.parameters(),lr=delta)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer,factor=0.95)
    prev_loss = 0.0
    counter = 0
    steps = 0
    while counter < 10:
        optimizer.zero_grad()
        loss = module()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(module.parameters(), 4.0*n**.5)
        optimizer.step()
        module.project_exp()
        scheduler.step(loss)
        loss = loss.item()
        steps += 1
        if abs(prev_loss - loss) < reltol/n*loss:
            counter += 1
        else:
            counter = 0
        prev_loss = loss
    
    return module().item()

def ferntala_s(W, normalized = True, reltol = 5e-3, delta = 0.05):
    """Computes the Fernique-Talagrand functional for a finite point set W
    on R^d with respect to the Euclidean distance. Uses the scipy built-in optimizer.
    - Input: the matrix W = (w_1,...,w_n) is d x n. 
             (optional) a flag for normalizing the functional (dividing by
                        the diameter of the set W (defaults to True)
             (optional) relative tolerance for the optimizer
             (optional) a step size delta > 0 for the optimizer
    - Output: a single positive value
    """
   
    n = W.size(1)
    # Use professional solver if problem is sufficiently small
    if (n <= 30) and _pyomo:
        return ft_pyomo(W)
    
    # Compute Euclidean distance matrix
    norms = torch.sum(torch.mul(W, W), 0).unsqueeze(0).repeat(n,1)
    distances = torch.sqrt(torch.abs(norms + norms.T - 2*W.T.mm(W)))
    # Normalize by max
    if normalized:
        distances = torch.clamp(distances, 0.0, 1.0)
    
    # Compute nearest-neighbour matrices
    distances, neighbours = torch.sort(distances, 1)
    distances = distances[:,1:n] - distances[:,0:n-1]
    neighbours = neighbours[:,:-1]
        
    # The Fernique-Talagrand integral
    def I_W(p):
        p_t = torch.from_numpy(p)
        p_t.requires_grad_(True)
        integrand = torch.sqrt(torch.abs(torch.log(torch.cumsum(p_t[neighbours],1))))
        loss = torch.sum(distances * integrand, 1).max()
        loss.backward()
        return loss.item(), p_t.grad.numpy()
    
    p_0 = np.random.dirichlet(np.ones(n))
    simplex_con1 = sciopt.LinearConstraint(np.eye(n), np.zeros(n), np.inf)
    simplex_con2 = sciopt.LinearConstraint(np.ones(n), np.ones(1),np.ones(1))
    result = sciopt.minimize(I_W, p_0, jac=True,constraints=(simplex_con1,simplex_con2),tol=1e-5)
    
    return result.fun
        
class FerniqueTalagrandModule(torch.nn.Module):
    def __init__(self, distances, neighbours, random_init = False):
        super(FerniqueTalagrandModule, self).__init__()
        self.distances = distances
        self.neighbours = neighbours
        n = distances.size(1) + 1
        self.p = torch.nn.Parameter(torch.from_numpy(np.random.dirichlet(np.ones(n))))
        self.p_store = self.p.clone().detach()
        
    @torch.no_grad()
    def get_integral_max(self, p):
        integrand = torch.sqrt(torch.abs(torch.log(torch.cumsum(p.clone()[self.neighbours],1))))
        vals = torch.sum(self.distances * integrand, 1)
        return vals.argmax()
        
    def integral(self, p):
        """The Fernique-Talagrand integral"""
        idx = self.get_integral_max(p)
        integrand = torch.sqrt(torch.abs(torch.log(torch.cumsum(p[self.neighbours[idx,:]],0))))
        val = torch.sum(self.distances[idx,:] * integrand)
        return val
    
    def forward(self):
        return self.integral(self.p)
    
    @torch.no_grad()
    def project(self):
        """Project the parameters back to the simplex via least squares (Wang, 2013)"""
        u = self.p.data.clone().sort(descending=True).values
        v = u + (1.0 - torch.cumsum(u,dim=0))/torch.arange(1,u.numel()+1)
        rho = torch.sum(v > 0)
        lam = 1.0/rho * (1.0 - torch.where(v > 0, u, torch.zeros(u.size())))
        torch.max(self.p.data.clone() + lam, torch.zeros(u.size()), out=self.p.data)        
    
    @torch.no_grad()
    def project_exp(self):
        """Project the parameters back to the simplex via exponential update"""
        self.p.add_(-self.p_store)
        self.p.exp_()
        self.p.mul_(self.p_store)
        self.p.div_(torch.sum(self.p))
        self.p_store = self.p.clone().detach()
        
    @torch.no_grad()
    def project_abs(self):
        """Project the parameters back to the simplex via absolute value"""
        self.p.abs_()
        self.p.div_(torch.sum(self.p))

def ft_pyomo(W):
    """Computes the Fernique-Talagrand functional for a finite point set W
    on R^d with respect to the Euclidean distance using Pyomo.
    - Input: the matrix W = (w_1,...,w_n) is d x n. 
    - Output: a single positive value
    """
    
    n = W.size(1)
    
    # Compute Euclidean distance matrix
    norms = torch.sum(torch.mul(W, W), 0).unsqueeze(0).repeat(n,1)
    D = torch.sqrt(torch.abs(norms + norms.T - 2*W.T.mm(W)))
    # Normalize by max
    D = D / torch.max(D)
    
    # Compute nearest-neighbour matrices
    D, S = torch.sort(D, 1)
    D = D[:,1:n] - D[:,0:n-1]
    S = S[:,:-1]
    D_p = D.tolist()
    S_p = S.tolist()
    
    # Find initial value
    p0 = torch.ones(n) / n
    integrand0 = torch.sqrt(torch.abs(torch.log(torch.cumsum(p0[S],1))))
    alpha0 = torch.sum(D * integrand0, 1).max().item()
    
    # The Fernique-Talagrand integral
    N = list(range(n))
    N1 = list(range(n-1))
    ft_model = pyo.ConcreteModel()
    ft_model.p = pyo.Var(N, initialize=1.0 / n, bounds=(0.0,1.0))
    ft_model.measure = pyo.Var(N, N1, bounds=(0.0,1.0))
    ft_model.alpha = pyo.Var(initialize=alpha0+1e-1)
    
    def meas(model, idx, idy):
        return model.measure[idx,idy] == sum([model.p[S_p[idx][l]] for l in range(idy+1)])
    
    def I_W(model,idc):
        integ = 0.0
        for k in N1:
            integ += D_p[idc][k] * pyo.sqrt(-pyo.log(model.measure[idc,k]))
        return model.alpha >= integ
    
    ft_model.simplex = pyo.Constraint(expr=sum([ft_model.p[k] for k in N])==1.0)
    ft_model.con1 = pyo.Constraint(N,N1,rule=meas)
    ft_model.con2 = pyo.Constraint(N, rule=I_W)
    ft_model.obj = pyo.Objective(expr=ft_model.alpha, sense=pyo.minimize)
    
    solver = pyo.SolverFactory('ipopt')
    solver.solve(ft_model)
    p = torch.Tensor([pyo.value(ft_model.p[i]) for i in N])
    integrand = torch.sqrt(torch.abs(torch.log(torch.cumsum(p[S],1))))
    
    return torch.sum(D * integrand, 1).max().item()

def _fernique_talagrand_v1(W, gamma = 0.05):
    """Computes the Fernique-Talagrand functional for a finite point set W
    on R^d with respect to the Euclidean distance. 
    - Input: the matrix W = (w_1,...,w_n) is d x n. 
             (optional) step size gamma > 0.
    - Output: a single positive value
    """
    
    n = W.size(1)
    
    # Compute Euclidean distance matrix
    norms = torch.sum(torch.mul(W, W), 0).unsqueeze(0).repeat(n,1)
    D = torch.sqrt(torch.abs(norms + norms.T - 2*W.T.mm(W)))
    # Normalize by max
    D = D / torch.max(D)
    
    # Compute nearest-neighbour matrices
    D, S = torch.sort(D, 1)
    D = D[:,1:n] - D[:,0:n-1]
    S = S[:,:-1]
    
    # The Fernique-Talagrand integral
    def I_W(p):
        integrand = torch.sqrt(torch.abs(torch.log(torch.cumsum(p[S],1))))
        return torch.sum(D * integrand, 1)
    
    ###################################################
    #    Optimize to find majorizing measure          #
    #    (probability distribution on points of W)    #
    #    via exponentiated gradient descent.          #
    ###################################################
    
    # Initial distribution drawn uniformly from simplex
    p = torch.from_numpy(np.random.dirichlet(np.ones(n)))
    # Exponentiated gradient descent
    gamma_step = gamma
    for epoch in range(2):
        for idx in range(500):
            p.requires_grad_(True)
            # Compute loss + gradient
            losses = I_W(p)
            ind = torch.argmax(losses).item()
            loss = losses[ind]
            g = torch.autograd.grad(loss, p)[0]
            # Update distribution
            p = p.detach()
            p = p * torch.exp(-torch.clamp(gamma_step * g, -4.0, 4.0))
            p = p / torch.sum(p)
        gamma_step *= gamma
    
    return torch.max(I_W(p)).item()
        

def _fernique_talagrand_v2(W, gamma = 0.05):
    """Computes the Fernique-Talagrand functional for a finite point set W
    on R^d with respect to the Euclidean distance. 
    - Input: the matrix W = (w_1,...,w_n) is d x n. 
             (optional) step size gamma > 0.
    - Output: a single positive value
    """
    
    n = W.size(1)
    
    # Compute Euclidean distance matrix
    norms = torch.sum(torch.mul(W, W), 0).unsqueeze(0).repeat(n,1)
    D = torch.sqrt(torch.abs(norms + norms.T - 2*W.T.mm(W)))
    # Normalize by max
    D = D / torch.max(D)
    
    # Compute nearest-neighbour matrices
    D, S = torch.sort(D, 1)
    D = D[:,1:n] - D[:,0:n-1]
    S = S[:,:-1]
    
    # The Fernique-Talagrand integral
    def I_W(p):
        integrand = torch.sqrt(torch.abs(torch.log(torch.cumsum(p[S],1))))
        return torch.sum(D * integrand, 1)
    
    # Initial distribution drawn uniformly from simplex
    p = torch.from_numpy(np.random.dirichlet(np.ones(n)))
    # Exponentiated gradient descent
    gamma_step = gamma
    loss_prev = I_W(p).max().item()
    curr_diff = I_W(p).max().item()
    for epoch in range(4):
        for idx in range(500):
            prev_diff = curr_diff
            p.requires_grad_(True)
            # Compute loss + gradient
            losses = I_W(p)
            ind = torch.argmax(losses).item()
            loss = losses[ind]
            g = torch.autograd.grad(loss, p)[0]
            # Update distribution
            p = p.detach()
            p = torch.abs(p - gamma_step * g)
            p = p / torch.sum(p)
            # Check convergence
            curr_diff = abs(loss-loss_prev)
            if curr_diff > prev_diff:
                break
        gamma_step *= 0.25
    
    return torch.max(I_W(p)).item()


if __name__ == '__main__':
    # Example test script. If all reported values are similar,
    # then the optimization is working correctly.
    
    
    W = torch.rand(10000,10)
    print('')
    print('Fernique-Talagrand functional for 10 points uniform\n'
          'in the 10000-dimensional unit cube:', ferntala(W),'\n')
    print('------------------------------------------------------\n')
    
    print('Fernique-Talagrand functional for 500 points from the\n'
          '100000-dimensional standard normal distribution\n')
    
    W = torch.randn(100000,500)
    for k in range(10):
        print('Trial', k, ': ', ferntala(W))