import numpy as np
import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import math
import copy
import pickle
import csv

import seaborn as sns
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn import preprocessing

from fernique_talagrand import ferntala
from math import log, exp
from random import sample

torch.set_num_threads(8)

rawdata = pd.read_csv('./datasets/winequality-white.csv',delimiter=';')
X_train, X_test, y_train, y_test = train_test_split(rawdata.loc[:,'fixed acidity':'alcohol'], rawdata.loc[:,'quality'], test_size=0.1, random_state=42)

means, stdev = X_train.mean(), X_train.std()
X_train = (X_train - means)/stdev
X_test = (X_test - means)/stdev
num_data, input_dim = X_train.shape
# Convert to PyTorch tensors
X_train = torch.tensor(X_train.values, dtype=torch.float64)
y_train = torch.unsqueeze(torch.tensor(y_train.values, dtype=torch.float64),1)
X_test = torch.tensor(X_test.values, dtype=torch.float64)
y_test = torch.unsqueeze(torch.tensor(y_test.values, dtype=torch.float64),1)

dataset = torch.utils.data.TensorDataset(X_train, y_train)
print('data =',X_train.shape[0], '; dims =', X_train.shape[1])

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

loss = nn.MSELoss()

def train(epoch, net, train_loader, optimizer, closure=False, keep_weights = False):
    prev_weights = torch.nn.utils.parameters_to_vector(net.parameters()).to('cpu').detach()
    net.train()
    l_track = 0.0
    idx = epoch * len(train_loader)
    weights = []
    if keep_weights:
        weights.append(prev_weights)
    for batch_idx, (data, target) in enumerate(train_loader):
        optimizer.zero_grad()
        data = data.to(device)
        target = target.to(device)
        if closure:
            def closure_fn():
                output = net(data)
                return loss(output, target)
        output = net(data)
        l = loss(output, target)
        l_track += l.item()
        if closure:
            optimizer.step(closure_fn)
        else:
            l.backward()
            optimizer.step()
        
        if keep_weights:
            # Record weights
            weights.append(torch.nn.utils.parameters_to_vector(net.parameters()).to('cpu').detach())
            
    return l_track / len(train_loader), weights

def test(net, test_loader):
    prev_weights = torch.nn.utils.parameters_to_vector(net.parameters()).detach()
    net.train()
    l_track = 0.0
    for batch_idx, (data, target) in enumerate(test_loader):
        data = data.to(device)
        target = target.to(device)
        output = net(data)
        l = loss(output, target)
        l_track += l.item()
            
    return l_track / len(test_loader)



hidden_units = 20
net = nn.Sequential(nn.Linear(X_train.shape[1], hidden_units), nn.ReLU(), nn.Linear(hidden_units, 1))
net = net.to('cuda')
state_dict = copy.deepcopy(net.state_dict())

test_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(X_test, y_test), batch_size=10)
ft = []
tl = []
num_repeats = 5

def convert_weights(weights, k):
    wt = sample(weights, k)
    M = torch.cat([w.unsqueeze(0) for w in wt], 0)
    return M.t()

with open('ft_loss_lg5.csv', 'w', newline='') as file:
    writer = csv.writer(file)
    writer.writerow(['ft','excess','lr','batch_size','wd','sigma'])
    for lr in (np.arange(5,15,1)/1000):
        for batch_size in [20, 60, 100]:
            for wd in [1e-4,5e-4,1e-3]:
                for sigma in [0.0, 0.05, 0.1]:
                    ft_val = 0.0
                    tl_val = 0.0
                    torch.manual_seed(0)
                    for idx in range(num_repeats):
                        noise_X = torch.randn_like(X_train)
                        dataset = torch.utils.data.TensorDataset(X_train+sigma*noise_X, y_train)
                        train_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True)
                        net.load_state_dict(copy.deepcopy(state_dict))
                        optimizer = torch.optim.SGD(net.parameters(), lr=lr, weight_decay=wd)
                        # initial weights
                        loss_val, _ = train(0, net, train_loader, optimizer, keep_weights=True)
                        # training
                        for epoch in range(10):
                            loss_val, _ = train(epoch, net, train_loader, optimizer)
                            #print('Epoch', epoch+1, ':', loss_val)
                            if(math.isnan(loss_val)):
                                print('Hit nan')
                                net.load_state_dict(copy.deepcopy(state_dict))
                                break
                        # collecting weights
                        weights = []
                        for epoch in range(10):
                            loss_val, weights_ep = train(epoch, net, train_loader, optimizer, keep_weights=True)
                            weights += weights_ep
                        # final weights
                        loss_val, weights_ep = train(epoch, net, train_loader, optimizer, keep_weights=True)
                        weights += weights_ep
                        ft_val += log(ferntala(convert_weights(weights, 50), rho=1.0))
                        tl_val += log(abs(loss_val - test(net, test_loader)))
                    ft_val = exp(ft_val / num_repeats)
                    tl_val = exp(tl_val / num_repeats)
                    print(ft_val,',',tl_val)
                    writer.writerow([ft_val, tl_val, lr, batch_size, wd, sigma])
                    ft.append(ft_val)
                    tl.append(tl_val)
            
